/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* sqlrext.h is required in all files using the RDM ODBC APIs */
#include "sqlrext.h"

/*! \file perf09.c
 * \brief Source code for the RDM Perf09 ODBC example
 */
 
/*!
 * \mainpage Perf09_ODBC
 * \author Daigoro F. Toyama, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf09 ODBC
 *
 * The Perf09 performance example creates 50,000 rows in a simple table
 * in an on-disk database.  It then spawns three threads.  The first one
 * updates rows in the table inside transactions with 1000 updates per
 * transaction.  The second thread attempts to read all the records to 
 * sum their contents.  The third thread just periodically reports the
 * results from the other two threads.  The reading thread uses standard
 * locks.  This can be compared against the performance in Perf10 where
 * the reading thread uses Read-Only Transactions.
 *
 * \section Perf09_db Database
 * The example utilizes the SQL DDL syntax and specifies an inmemory database
 * with a single table (simple).  The table has a single column (int_col).
 * There are no index definitions for this database.
 * The SQL DDL commands for creating the database are executed in the 
 * initialization function, but the commands used to create the database can
 * also be found in the file perf09_db.sdl.
 *
 * \include perf09_db.sdl
 *
 * \section RDM ODBC Functions Used
 * The following RDM ODBC functions are illustrated:
 *   * SQLAllocHandle/SQLFreeHandle
 *   * SQLGetDiagRec
 *   * SQLConnect/SQLDisconnect
 *   * SQLExecDirect
 *   * SQLEndTran
 *   * SQLPrepare
 *   * SQLBindParameter
 *   * SQLExecute
 *   * SQLBindCol
 *   * SQLFetch
 *   * SQLEndTran
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */
#define NumRows 50000

static const int32_t transactionSize = 1000;

/*! \brief Create a test database
 *
 */
static SQLRETURN create_database(
    SQLHDBC hDbc)
{
    SQLRETURN sqlError;
    SQLHSTMT  hStmt;
    int16_t   ii;
    SQLRETURN dai;

    const char *cmds[] = {
        "CREATE DATABASE perf09_db",
        "CREATE TABLE simple (id INTEGER PRIMARY KEY, int_col INTEGER)",
        "COMMIT"
    };

    (void) SQLAllocHandle(SQL_HANDLE_STMT, hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    dai = SQLExecDirect(hStmt, (SQLCHAR *) "DROP DATABASE perf09_db", SQL_NTS);
    dai = SQLEndTran(SQL_HANDLE_DBC, hDbc, SQL_COMMIT);

    for (ii = 0; ii < sizeof(cmds) / sizeof(cmds[0]); ii++)
    {
        sqlError = SQLExecDirect(hStmt, (SQLCHAR *) cmds[ii], SQL_NTS);
        if (!SQL_SUCCEEDED(sqlError))
            break;
    }

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hStmt);
    return SQL_SUCCESS;
}

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void display_elapsed_time(
    const char *msg,         /**< [in] Message to be display along with the elapsed time */
    uint64_t    start_time,  /**< [in] Start of elapsed time to display */
    uint64_t    end_time)    /**< [in] End of elapsed */
{
    uint64_t elapsed_time = end_time - start_time;
    printf("%s" UINT64_SPEC " millisecs\n", msg, elapsed_time/1000000);
}

/*! \brief Display a RDM ODBC Error message
 *
 * This function retrieves error information from RDM ODBC and
 * displays it and the passed in message to stderr.
 */
void display_error_msg(
    SQLSMALLINT handle_type,  /**< [in] Type of ODBC Handle that triggered the error */
    SQLHANDLE   handle,       /**< [in] ODBC Handle that triggered the error */
    const char *msg           /**< [in] Message to be display along with ODBC error */
    )
{
    SQLCHAR SQLState[6];
    SQLINTEGER NativeError;
    SQLCHAR MessageText[100];
    SQLSMALLINT Len;

    SQLGetDiagRec(handle_type, handle, 1, SQLState, &NativeError, MessageText, sizeof(MessageText), &Len);
    
    fprintf(stderr, "%s: %s - %s (%d)\n", msg, SQLState, MessageText, NativeError);
}

/*! \brief Insert 50,000 simple records into the perf09 database
 *
 * This function adds 50,000 simple records into the inmemory perf database.  The
 * simple table contains only a single integer column with no index.  The function will
 * calculate and display the total time elapsed while inserting the records
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN add_records(
    SQLHANDLE hDBC)         /**< [in] RDM ODBC Connection handle */
{
    SQLHANDLE hSTMT;
    int32_t   id;
    int32_t   ii;
    SQLRETURN sqlError = SQL_SUCCESS;

    /* allocates insert statement */
    sqlError = SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, hDBC, "Error has occured allocating insert statement");
        return sqlError;
    }

    /* Prepare insert statement */
    sqlError = SQLPrepare(hSTMT, (SQLCHAR *) "INSERT INTO simple VALUES ?, 0", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occured preparing the insert statement");
        SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    /* Bind the insert value to variable location */
    sqlError = SQLBindParameter(hSTMT, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, &id, 0, NULL);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occurred binding the insert statement parameter");
        SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    /* Loop to add 50,000 new simple record */
    for (ii = 0; ii < NumRows; ii++)
    {
        /* Set values for the new record */
        id = ii;

        /* Add the record to the database */
        sqlError = SQLExecute(hSTMT);
        if (!SQL_SUCCEEDED(sqlError))
        {            
            display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occurred exeucting the insert statement");
            break;
        }
    }

    (void) SQLEndTran(SQL_HANDLE_DBC, hDBC, SQL_COMMIT);
    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);

    return sqlError;
}

/*! \brief Initialize the RDM ODBC library for use in the perf09 example
 *
 * This function 
 *   * Initializes the RDM Transactional File Server (TFS) to use the 
 *     single-user non-transactional STANDALONE implementation
 *   * Creates and intializes the simple database
 *   * Creates and connection an ODBC connection handle
 *   * Opens the "perf09" database in exclusive ('x') mode.
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN initialize(
    SQLHANDLE    *hENV,    /**< [out] Pointer to the environment handle to be initialized */
    SQLHANDLE    *hDBC)    /**< [out] Pointer to the connection handle to be initialized */
{
    SQLRETURN sqlError;

    *hENV = NULL;
    *hDBC = NULL;

    /* Allocate the environment handle */
    sqlError = SQLAllocHandle(SQL_HANDLE_ENV, NULL, hENV);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Error allocating an environment handle\n");
        return sqlError;
    }

    /* Allocate the connection handle */
    sqlError = SQLAllocHandle(SQL_HANDLE_DBC, *hENV, hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {       
        display_error_msg(SQL_HANDLE_ENV, *hENV, "Error allocating a connection handle");
        return sqlError;
    }

    /* Initialize the TFS to use the DIREC-LINK configuration (TFST) */
    sqlError = SQLSetConnectAttr(*hDBC, SQL_ATTR_RDM_TFSINIT_TYPE,
            (SQLPOINTER) SQL_TFSTYPE_TFS, 0);
    if (!SQL_SUCCEEDED(sqlError))
    {       
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error setting the TFS type");
        return sqlError;
    }

    /* Connect to the RDM SQL Server */
    sqlError = SQLConnect(*hDBC,
                            (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error calling SQLConnect");
        return sqlError;
    }

    /* Create the "perf09_db" database */
    sqlError = create_database(*hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error creating a database");
        return sqlError;
    }

    (void) add_records(*hDBC);

    (void) SQLDisconnect(*hDBC);

    /* Display information */
    printf("Initializing with DIRECT-LINK configuration\n");
    printf("\tOne, simple record type.\n");
    printf("\tOn disk\n");
    printf("\tOpened in shared mode\n");
    printf("\t\t\tHit <Enter> to exit the example\n");

    return sqlError;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM connection
 * handle used in the perf09 example.  It also terminates the STANDALONE TFS
 * used in the example. 
 */
static void cleanup(
    SQLHANDLE     hENV,    /**< [in] Environment handle to be cleaned up */
    SQLHANDLE     hDBC)    /**< [in] Connection handle to be cleaned up */
{
    if (hDBC)
    {
        /* Close any open databases */
        (void) SQLDisconnect(hDBC);

        /* Free the connection handle */
        (void) SQLFreeHandle(SQL_HANDLE_DBC, hDBC);
    }

    /* Free the environment handle */
    if (hENV)
        (void) SQLFreeHandle(SQL_HANDLE_ENV, hENV);
}

static int16_t quit = 0;
static int64_t rowsUpdated = 0;
static int64_t queriesExecuted = 0;

/*! \brief Thread function for updating the rows
 *
 * The function updates the rows stored in the simple table
 */
static void *updater_thread(
    void *ctx)
{
    SQLHENV  hENV;
    SQLHDBC  hDBC;
    SQLHSTMT hSTMT;
    int32_t  minid_parm;
    int32_t  minid;

    printf("Updating rows in transactions of size %d...\n", transactionSize);

    /* Create a connection for this thread */
    (void) SQLAllocHandle(SQL_HANDLE_ENV, NULL, &hENV);
    (void) SQLAllocHandle(SQL_HANDLE_DBC, hENV, &hDBC);
    (void) SQLConnect(hDBC, (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS);

    (void) SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);

    (void) SQLExecDirect(hSTMT, (SQLCHAR *) "OPEN perf09_db", SQL_NTS);
    (void) SQLPrepare(hSTMT, (SQLCHAR *) "UPDATE simple SET int_col = int_col + 1 WHERE id >= ? AND id < ? + 1000", SQL_NTS);
    (void) SQLBindParameter(hSTMT, 1, SQL_PARAM_INPUT, SQL_C_SLONG,
            SQL_INTEGER, 0, 0, &minid_parm, 0, NULL);
    (void) SQLBindParameter(hSTMT, 2, SQL_PARAM_INPUT, SQL_C_SLONG,
            SQL_INTEGER, 0, 0, &minid_parm, 0, NULL);

    minid = 0;

    while (!quit)
    {
        if (minid >= NumRows)
            minid = 0;

        minid_parm = minid;
        (void) SQLExecute(hSTMT);

        (void) SQLEndTran(SQL_HANDLE_DBC, hDBC, SQL_COMMIT);

        rowsUpdated += transactionSize;
        minid += transactionSize;
    }

    printf("Ending updates\n");

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);

    (void) SQLDisconnect(hDBC);
    (void) SQLFreeHandle(SQL_HANDLE_DBC, hDBC);
    (void) SQLFreeHandle(SQL_HANDLE_ENV, hENV);

    return NULL;
}

/*! \brief Thread function for reading all the records to sum their contents
 *
 * The function reads all the records to sum thier contents.
 */
static void *reader_thread(
    void *ctx)
{
    SQLHENV  hENV;
    SQLHDBC  hDBC;
    SQLHSTMT hSTMT;
    char     commandText[80];

    /* Create a connection for this thread */
    (void) SQLAllocHandle(SQL_HANDLE_ENV, NULL, &hENV);
    (void) SQLAllocHandle(SQL_HANDLE_DBC, hENV, &hDBC);
    (void) SQLConnect(hDBC, (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS);

    (void) SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);
    (void) SQLExecDirect(hSTMT, (SQLCHAR *) "OPEN perf09_db", SQL_NTS);

    sprintf(commandText, "SELECT SUM(int_col) FROM simple");

    printf("Reading the contents of the table...\n");

    (void) SQLPrepare(hSTMT, (SQLCHAR *) commandText, SQL_NTS);

    while (!quit)
    {
        (void) SQLExecute(hSTMT);
        (void) SQLFreeStmt(hSTMT, SQL_CLOSE);
        queriesExecuted++;
    }

    printf("Ending reading\n");
        
    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);

    (void) SQLDisconnect(hDBC);
    (void) SQLFreeHandle(SQL_HANDLE_DBC, hDBC);
    (void) SQLFreeHandle(SQL_HANDLE_ENV, hENV);

    return NULL;
}

/*! \brief Thread function for reporting the results
 *
 * The function periodically reports results from the other two threads.
 */
static void *reporter_thread(
    void *ctx)
{
    int32_t secs;

    for (secs = 1; ; secs++)
    {
        sleep(1);

        if (quit)
            break;

        printf("Updated %ld rows (%ld/sec) and executed %ld queries (%.2lf/sec) in %d secs\n", rowsUpdated, (rowsUpdated / secs), queriesExecuted, ((double) queriesExecuted / secs), secs);
    }

    return NULL;
}

/*! \brief Main function for perf09 example
 *
 * The function initializes the RDM ODBC environment and runs the create, read,
 * update, and delete operations.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    SQLRETURN sqlError;
    SQLHANDLE hDBC;
    SQLHANDLE hENV;
    pthread_t updater_tid;
    pthread_t reader_tid;
    pthread_t reporter_tid;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf09 ODBC"));

    /* Initialize the TFS, task, and open/initialize the database */
    sqlError = initialize(&hENV, &hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    (void) pthread_create(&updater_tid, NULL, &updater_thread, NULL);
    (void) pthread_create(&reader_tid, NULL, &reader_thread, NULL);
    (void) pthread_create(&reporter_tid, NULL, &reporter_thread, NULL); 

    /* Wait for user to hit <Enter> */
    getchar();

    /* Indicate to the threads that they should shut down */
    quit = 1;

    /* Wait for each thread to finish */
    pthread_join(updater_tid, NULL);
    pthread_join(reader_tid, NULL);
    pthread_join(reporter_tid, NULL);

    cleanup(hENV, hDBC);

    printf("\nPress Enter to terminate");
    getchar();

    return EXIT_SUCCESS;
}
